from collections import namedtuple

import numpy as np
import pandas as pd
import pytest

from toarstats.metrics.input_checks import (
    check_crops,
    check_data,
    check_min_data_capture,
    check_index,
    check_input_parameters,
    check_metadata,
    check_required_parameters,
    check_sampling,
    check_seasons,
    check_station_climatic_zone,
    check_station_latitude,
    check_station_longitude,
    check_statistics,
    check_values,
    from_pandas,
    is_correct_type,
    is_empty,
    is_in_range,
    to_collection
)


class TestIsCorrectType:
    def test_is_correct_type_non_collection_single_type(self):
        assert is_correct_type(True, bool)
        assert is_correct_type("string", str)
        assert is_correct_type(3, int)
        assert is_correct_type(7.12, float)
        assert not is_correct_type(True, float)
        assert not is_correct_type("string", bool)
        assert not is_correct_type(3, str)
        assert not is_correct_type(7.12, int)

    def test_is_correct_type_non_collection_multiple_types(self):
        assert is_correct_type(True, (bool, str))
        assert is_correct_type("string", (str, int))
        assert is_correct_type(3, (int, float))
        assert is_correct_type(7.12, (float, bool))
        assert not is_correct_type(True, (float, str))
        assert not is_correct_type("string", (bool, int))
        assert not is_correct_type(3, (str, float))
        assert not is_correct_type(7.12, (int, bool))

    def test_is_correct_type_collection_single_type(self):
        assert is_correct_type([True], list, bool)
        assert is_correct_type(["string"], list, str)
        assert is_correct_type([3], list, int)
        assert is_correct_type([7.12], list, float)
        assert not is_correct_type([True], list, float)
        assert not is_correct_type(["string"], list, bool)
        assert not is_correct_type([3], list, str)
        assert not is_correct_type([7.12], list, int)

    def test_is_correct_type_collection_multiple_types(self):
        assert is_correct_type([True], list, (bool, str))
        assert is_correct_type(["string"], list, (str, int))
        assert is_correct_type([3], list, (int, float))
        assert is_correct_type([7.12], list, (float, bool))
        assert not is_correct_type([True], list, (float, str))
        assert not is_correct_type(["string"], list, (bool, int))
        assert not is_correct_type([3], list, (str, float))
        assert not is_correct_type([7.12], list, (int, bool))

    def test_is_correct_type_collection_multiple_values_multiple_types(self):
        assert is_correct_type(["string", "another string", 3, 4],
                               (list, set, tuple), (str, int))
        assert not is_correct_type(["string", "another string", 3, 4],
                                   (list, set, tuple), (str, float))


class TestCheckSampling:
    def test_check_sampling_wrong_type(self):
        error_msg = "The sampling parameter is required and must be a string"
        for argument in (3.14, True, 100, {"A", "B", "C"}):
            with pytest.raises(TypeError) as excinfo:
                check_sampling(argument)
            assert error_msg == str(excinfo.value)

    def test_check_sampling_wrong_value(self):
        error_msg = "The sampling parameter must be one of"
        for argument in ("abc", "test", "not right"):
            with pytest.raises(ValueError) as excinfo:
                check_sampling(argument)
            assert str(excinfo.value).startswith(error_msg)

    def test_check_sampling_right_value(self):
        assert check_sampling("daily") == "daily"
        assert check_sampling("  vegseason    ") == "vegseason"
        assert check_sampling("annual  ") == "annual"
        assert check_sampling("     monthly") == "monthly"


class TestIsEmpty:
    def test_is_empty_collections(self):
        assert is_empty([])
        assert is_empty(tuple())
        assert is_empty(set())
        assert not is_empty([1, 2])
        assert not is_empty((1, 2))
        assert not is_empty({1, 2})

    def test_is_empty_string(self):
        assert is_empty("")
        assert is_empty("\t\n")
        assert is_empty("         ")
        assert not is_empty("string")
        assert not is_empty("\tstring\n")
        assert not is_empty("    string    ")


class TestToCollection:
    def test_to_collection_collections(self):
        assert to_collection(["test"]) == ["test"]
        assert to_collection(("test",)) == ("test",)
        assert to_collection({"test"}) == {"test"}

    def test_to_collection_string(self):
        assert to_collection("test") == ["test"]
        assert to_collection("string") == ["string"]


class TestCheckStatistics:
    def test_check_statistics_wrong_type(self):
        error_msg = ("The statistics parameter is required and must either be"
                     " a single string or a collection of strings in the form"
                     " of a list, set or tuple")
        for argument in (3.14, True, 100, ["A", 3.1], ("A", 12), {"A", 1.2}):
            with pytest.raises(TypeError) as excinfo:
                check_statistics(argument)
            assert error_msg == str(excinfo.value)

    def test_check_statistics_empty(self):
        error_msg = "No statistics given"
        for argument in ([], tuple(), set(), "", "   "):
            with pytest.raises(ValueError) as excinfo:
                check_statistics(argument)
            assert error_msg == str(excinfo.value)

    def test_check_statistics_wrong_value(self):
        error_msg = "Test is not a recognized statistic, available are"
        with pytest.raises(ValueError) as excinfo:
            check_statistics(["dma8epax", "Test", "p95"])
        assert str(excinfo.value).startswith(error_msg)

    def test_check_statistics_diurnal_cycle_not_alone(self):
        error_msg = "Diurnal cycle can only be evaluated alone"
        with pytest.raises(ValueError) as excinfo:
            check_statistics(["dma8epax", "diurnal_cycle", "p95"])
        assert error_msg == str(excinfo.value)

    def test_check_statistics_right_values(self):
        assert check_statistics("diurnal_cycle") == ["diurnal_cycle"]
        assert check_statistics(["diurnal_cycle"]) == ["diurnal_cycle"]
        assert check_statistics(("p95", "somo35")) == ["p95", "somo35"]
        assert check_statistics(("p95", "p95", "somo35")) == ["p95", "somo35"]
        assert check_statistics({"p95"}) == ["p95"]


class TestFromPandas:
    def test_from_pandas_series(self):
        index = pd.date_range(start="2000", periods=3, freq="H")
        out_index, out_value = from_pandas(pd.Series([1., 2., 3.], index))
        pd.testing.assert_index_equal(out_index, index)
        np.testing.assert_array_equal([1., 2., 3.], out_value)

    def test_from_pandas_data_frame(self):
        index = pd.date_range(start="2000", periods=2, freq="H")
        frame1 = pd.DataFrame({"values": [3., 2.], "value": [1., 2.]}, index)
        frame2 = pd.DataFrame({"values": [3., 2.], "other": [1., 2.]}, index)
        frame3 = pd.DataFrame({"one": [5., 2.], "two": [1., 2.]}, index)
        frame4 = pd.DataFrame(index=index)
        frame5 = pd.DataFrame()
        out_index1, out_value1 = from_pandas(frame1)
        out_index2, out_value2 = from_pandas(frame2)
        out_index3, out_value3 = from_pandas(frame3)
        out_index4, out_value4 = from_pandas(frame4)
        out_index5, out_value5 = from_pandas(frame5)
        pd.testing.assert_index_equal(out_index1, index)
        np.testing.assert_array_equal([1., 2.], out_value1)
        pd.testing.assert_index_equal(out_index2, index)
        np.testing.assert_array_equal([3., 2.], out_value2)
        pd.testing.assert_index_equal(out_index3, index)
        np.testing.assert_array_equal([5., 2.], out_value3)
        pd.testing.assert_index_equal(out_index4, index)
        np.testing.assert_array_equal([], out_value4)
        pd.testing.assert_index_equal(out_index5, pd.RangeIndex(0))
        np.testing.assert_array_equal([], out_value5)


class TestCheckIndex:
    def test_check_index_wrong_type(self):
        error_msg = "No datetime index given"
        with pytest.raises(TypeError) as excinfo:
            check_index(["2000-01-01 00:00", "2000-01-01 01:00"])
        assert error_msg == str(excinfo.value)

    def test_check_index_wrong_value(self):
        error_msg = ("The datetime index may not be empty, have duplicates or"
                     " null values")
        index1 = pd.DatetimeIndex([])
        index2 = pd.DatetimeIndex(["2000-01-01 00:00", "2000-01-01 00:00"])
        index3 = pd.DatetimeIndex(["2000-01-01 00:00", "NaT"])
        for index in (index1, index2, index3):
            with pytest.raises(ValueError) as excinfo:
                check_index(index)
            assert error_msg == str(excinfo.value)

    def test_check_index_correct_index(self):
        index = pd.DatetimeIndex(["2000-01-01 00:00", "2000-01-01 01:00"])
        assert check_index(index) is None


class TestCheckValues:
    def test_check_values_wrong_type(self):
        error_msg = ("Values must be given in a one-dimensional float or int"
                     " array")
        values1 = [1., 2., 3.]
        values2 = np.array([[1.], [2.], [3.]])
        values3 = np.array(["Hello", "World", 1, 2.])
        for argument in (values1, values2, values3):
            with pytest.raises(TypeError) as excinfo:
                check_values(argument)
            assert error_msg == str(excinfo.value)

    def test_check_values_empty(self):
        error_msg = "No values given"
        values1 = np.array([])
        values2 = np.array([np.nan, np.nan])
        for argument in (values1, values2):
            with pytest.raises(ValueError) as excinfo:
                check_values(argument)
            assert error_msg == str(excinfo.value)

    def test_check_values_right_values(self):
        values1 = np.array([1, 2., 3, 4.])
        values2 = np.array([1., np.nan, 3, 4.])
        values3 = np.array([1, 2, 3])
        for argument in (values1, values2, values3):
            assert check_values(argument) is None


class TestCheckData:
    def test_check_data_wrong_type_data(self):
        error_msg = "The data parameter must be a data frame or series"
        for argument in ("Test", 1.2, 50, True):
            with pytest.raises(TypeError) as excinfo:
                check_data(argument, None, None)
            assert error_msg == str(excinfo.value)

    def test_check_data_wrong_type_index(self):
        error_msg = "Index(...) must be called with a collection of some kind"
        for argument in ("Test", 1.2, 50, True):
            with pytest.raises(TypeError) as excinfo:
                check_data(None, argument, 1)
            assert str(excinfo.value).startswith(error_msg)

    def test_check_data_different_lengths(self):
        error_msg = "Length of values (3) does not match length of index (1)"
        with pytest.raises(ValueError) as excinfo:
            check_data(None, pd.DatetimeIndex(["2000"]), [1, 2, 3])
        assert error_msg == str(excinfo.value)

    def test_check_data_correct_data(self):
        index = pd.date_range(
            start="2000", periods=3, freq="H", tz="Europe/Berlin"
        )
        values = np.array([5., 6.1, 7])
        expected = pd.Series(values, index=index.tz_localize(None))
        pd.testing.assert_series_equal(
            check_data(pd.DataFrame(values, index=index), None, None), expected
        )
        pd.testing.assert_series_equal(
            check_data(pd.Series(values, index=index), None, None), expected
        )
        pd.testing.assert_series_equal(
            check_data(None, index, values), expected
        )


class TestIsInRange:
    def test_is_in_range_float(self):
        assert is_in_range(0.1, 0, 1)
        assert is_in_range(0.9, 0, 1)
        assert is_in_range(0., 0, 1)
        assert is_in_range(1., 0, 1)
        assert is_in_range(0., 0., 1.)
        assert is_in_range(1., 0., 1.)
        assert not is_in_range(-0.1, 0, 1)
        assert not is_in_range(1.1, 0, 1)

    def test_is_in_range_int(self):
        assert is_in_range(3, 1, 7)
        assert is_in_range(5, 1, 7)
        assert is_in_range(1, 1, 7)
        assert is_in_range(7, 1, 7)
        assert not is_in_range(0, 1, 7)
        assert not is_in_range(8, 1, 7)


class TestCheckStationLatitude:
    def test_check_station_latitude_wrong_type(self):
        error_msg = "The station latitude must be a float or int value"
        for argument in ("Test", [], {1, 2, 3}, (True,)):
            with pytest.raises(TypeError) as excinfo:
                check_station_latitude(argument)
            assert error_msg == str(excinfo.value)

    def test_check_station_latitude_wrong_range(self):
        error_msg = "The station latitude must be between -90 and 90"
        for argument in (-190., 2000, 90.00001, 91):
            with pytest.raises(ValueError) as excinfo:
                check_station_latitude(argument)
            assert error_msg == str(excinfo.value)

    def test_check_station_latitude_right_value(self):
        for argument in (-80, 0, 1.2, 20.23):
            assert check_station_latitude(argument) is None


class TestCheckStationLongitude:
    def test_check_station_longitude_wrong_type(self):
        error_msg = "The station longitude must be a float or int value"
        for argument in ("Test", [], {1, 2, 3}, (True,)):
            with pytest.raises(TypeError) as excinfo:
                check_station_longitude(argument)
            assert error_msg == str(excinfo.value)

    def test_check_station_longitude_wrong_range(self):
        error_msg = "The station longitude must be between -180 and 180"
        for argument in (-190., 2000, 180.00001, 181):
            with pytest.raises(ValueError) as excinfo:
                check_station_longitude(argument)
            assert error_msg == str(excinfo.value)

    def test_check_station_longitude_right_value(self):
        for argument in (-80, 0, 1.2, 20.23):
            assert check_station_longitude(argument) is None


class TestCheckStationClimaticZone:
    def test_check_station_climatic_zone_wrong_type(self):
        error_msg = ("The station climatic zone must be an int or string"
                     " contained in")
        for argument in (3.1, [], {1, 2, 3}, (True,)):
            with pytest.raises(TypeError) as excinfo:
                check_station_climatic_zone(argument)
            assert str(excinfo.value).startswith(error_msg)

    def test_check_station_climatic_zone_wrong_int(self):
        error_msg = "The station climatic zone must be containted in"
        for argument in (-2, 13, 14):
            with pytest.raises(ValueError) as excinfo:
                check_station_climatic_zone(argument)
            assert str(excinfo.value).startswith(error_msg)

    def test_check_station_climatic_zone_right_int(self):
        assert check_station_climatic_zone(-1) == "undefined"
        assert check_station_climatic_zone(0) == "unclassified"
        assert check_station_climatic_zone(1) == "warm temperate moist"
        assert check_station_climatic_zone(2) == "warm temperate dry"
        assert check_station_climatic_zone(3) == "cool temperate moist"
        assert check_station_climatic_zone(4) == "cool temperate dry"
        assert check_station_climatic_zone(5) == "polar moist"
        assert check_station_climatic_zone(6) == "polar dry"
        assert check_station_climatic_zone(7) == "boreal moist"
        assert check_station_climatic_zone(8) == "boreal dry"
        assert check_station_climatic_zone(9) == "tropical montane"
        assert check_station_climatic_zone(10) == "tropical wet"
        assert check_station_climatic_zone(11) == "tropical moist"
        assert check_station_climatic_zone(12) == "tropical dry"

    def test_check_station_climatic_zone_wrong_str(self):
        error_msg = "The station climatic zone must be containted in"
        for argument in ("Test", "climatic", "zone"):
            with pytest.raises(ValueError) as excinfo:
                check_station_climatic_zone(argument)
            assert str(excinfo.value).startswith(error_msg)

    def test_check_station_climatic_zone_right_str(self):
        assert (check_station_climatic_zone("   warm temperate moist  ")
                == "warm temperate moist")
        assert check_station_climatic_zone("polar dry") == "polar dry"
        assert check_station_climatic_zone("  boreal dry") == "boreal dry"


class TestCheckMetadata:
    def test_check_metadata_wrong_type_metadata(self):
        error_msg = ("The metadata parameter must be a dictionary or named"
                     " tuple")
        for argument in (3.1, 1, "Test", True, (1, 2, 3)):
            with pytest.raises(TypeError) as excinfo:
                check_metadata(argument, None, None, None)
            assert error_msg == str(excinfo.value)

    def test_check_metadata_wrong_keys(self):
        error_msg = "The metadata parameter can only have keys in"
        metadata1 = {"station_lat": 1, "wrong_key": 2}
        metadata2 = {"station_lat": 1, "station_lon": 2, "wrong_key": 3,
                     "station_climatic_zone": "polar dry"}
        metadata3 = namedtuple("Metadata", "station_lat, wrong_key")(1, 2)
        metadata4 = namedtuple(
            "Metadata",
            "station_lat, station_lon, wrong_key, station_climatic_zone"
        )(1, 2, 3, "polar dry")
        for argument in (metadata1, metadata2, metadata3, metadata4):
            with pytest.raises(KeyError) as excinfo:
                check_metadata(argument, None, None, None)
            assert str(excinfo.value)[1:].startswith(error_msg)

    def test_check_metadata_right_values(self):
        assert check_metadata(None, None, None, None) is None
        assert (check_metadata(None, 1, None, None)
                == namedtuple("Metadata", "station_lat")(1))
        assert (check_metadata(None, None, 2, None)
                == namedtuple("Metadata", "station_lon")(2))
        assert (check_metadata(None, None, None, "boreal moist")
                == namedtuple("Metadata",
                              "station_climatic_zone")("boreal moist"))
        metadata = namedtuple(
            "Metadata", "station_lat, station_lon, station_climatic_zone"
        )(1, 2, "polar dry")
        assert (check_metadata(
                    {"station_lat": 1, "station_lon": 2,
                     "station_climatic_zone": "polar dry"}, None, None, None
                ) == metadata)
        assert check_metadata(metadata, 100, 25, "boreal dry") == metadata
        assert check_metadata(None, 1, 2, "polar dry") == metadata


class TestCheckSeasons:
    def test_check_seasons_wrong_type(self):
        error_msg = ("The seasons parameter must either be a single string or"
                     " a collection of strings in the form of a list, set or"
                     " tuple")
        for argument in (3.1, 1, ["a", 1.1, "c"], True, (1, 2, 3)):
            with pytest.raises(TypeError) as excinfo:
                check_seasons(argument)
            assert error_msg == str(excinfo.value)

    def test_check_seasons_empty(self):
        error_msg = "No seasons given"
        for argument in ([], tuple(), set(), "", "   "):
            with pytest.raises(ValueError) as excinfo:
                check_seasons(argument)
            assert error_msg == str(excinfo.value)

    def test_check_seasons_wrong_value(self):
        error_msg = "Test is not a recognized season, available are"
        with pytest.raises(ValueError) as excinfo:
            check_seasons(["DJF", "Test", "MAM"])
        assert str(excinfo.value).startswith(error_msg)

    def test_check_seasons_right_values(self):
        assert check_seasons(None) is None
        assert check_seasons("DJF") == ["DJF"]
        assert check_seasons(["DJF"]) == ["DJF"]
        assert check_seasons(("MAM", "SON")) == ["MAM", "SON"]
        assert check_seasons(("MAM", "MAM", "SON")) == ["MAM", "SON"]
        assert check_seasons({"JJA"}) == ["JJA"]


class TestCheckCrops:
    def test_check_crops_wrong_type(self):
        error_msg = ("The crops parameter must either be a single string or a"
                     " collection of strings in the form of a list, set or"
                     " tuple")
        for argument in (3.1, 1, ["a", 1.1, "c"], True, (1, 2, 3)):
            with pytest.raises(TypeError) as excinfo:
                check_crops(argument)
            assert error_msg == str(excinfo.value)

    def test_check_crops_empty(self):
        error_msg = "No crops given"
        for argument in ([], tuple(), set(), "", "   "):
            with pytest.raises(ValueError) as excinfo:
                check_crops(argument)
            assert error_msg == str(excinfo.value)

    def test_check_crops_wrong_value(self):
        error_msg = "Test is not a recognized crop, available are"
        with pytest.raises(ValueError) as excinfo:
            check_crops(["wheat", "Test", "rice"])
        assert str(excinfo.value).startswith(error_msg)

    def test_check_crops_right_values(self):
        assert check_crops(None) is None
        assert check_crops("wheat") == ["wheat"]
        assert check_crops(["wheat"]) == ["wheat"]
        assert check_crops(("wheat", "rice")) == ["wheat", "rice"]
        assert check_crops(("wheat", "wheat", "rice")) == ["wheat", "rice"]
        assert check_crops({"wheat"}) == ["wheat"]


class TestCheckMinDataCapture:
    def test_check_min_data_capture_wrong_type(self):
        error_msg = ("The min data capture parameter must be a float or int"
                     " value")
        for argument in ("abc", {1, 2}, ["a", 1.1, "c"], "True", (1, 2, 3)):
            with pytest.raises(TypeError) as excinfo:
                check_min_data_capture(argument)
            assert error_msg == str(excinfo.value)

    def test_check_min_data_capture_wrong_value(self):
        error_msg = "The min data capture parameter must be between 0 and 1"
        for argument in (-0.1, 1.1, 2, 5):
            with pytest.raises(ValueError) as excinfo:
                check_min_data_capture(argument)
            assert error_msg == str(excinfo.value)

    def test_check_min_data_capture_right_values(self):
        assert check_min_data_capture(None) is None
        assert check_min_data_capture(0) == 0
        assert check_min_data_capture(0.) == 0.
        assert check_min_data_capture(1) == 1
        assert check_min_data_capture(1.) == 1.
        assert check_min_data_capture(0.75) == 0.75


class TestCheckRequiredParameters:
    metadata = namedtuple(
        "Metadata", "station_lat, station_lon, station_climatic_zone"
    )(50.906389, 6.403889, "cool temperate moist")
    Required = namedtuple(
        "Required",
        "station_lat, station_lon, station_climatic_zone, seasons, crops"
    )

    def test_check_required_parameters_statistics_with_lat_lon(self):
        expected = self.Required(True, True, False, False, False)
        assert check_required_parameters(
            "annual", ["median", "dark_aot40", "p90"], self.metadata, None
        ) == expected

    def test_check_required_parameters_statistics_summer_sampling(self):
        expected = self.Required(True, False, False, False, False)
        assert check_required_parameters(
            "xsummer", ["median", "p90"], self.metadata, None
        ) == expected

    def test_check_required_parameters_statistics_seasonal_sampling(self):
        expected = self.Required(False, False, False, True, False)
        assert check_required_parameters(
            "seasonal", ["median", "p90"], self.metadata, None
        ) == expected

    def test_check_required_parameters_statistics_seasonal_with_crops(self):
        expected = self.Required(True, False, True, True, True)
        assert check_required_parameters(
            "seasonal", ["median", "w126", "p90"], self.metadata, None
        ) == expected

    def test_check_required_parameters_statistics_vegseason_sampling(self):
        expected = self.Required(True, False, True, True, True)
        assert check_required_parameters(
            "vegseason", ["median", "p90"], self.metadata, None
        ) == expected

    def test_check_required_parameters_missing_metadata(self):
        error_msg = ("The station latitude is required and must be a float or"
                     " int value between -90 and 90")
        tmp_dict = self.metadata._asdict()
        del tmp_dict["station_lat"]
        with pytest.raises(TypeError) as excinfo:
            check_required_parameters(
                "seasonal", ["median", "dark_aot40", "p90"],
                namedtuple("Metadata", tmp_dict)(**tmp_dict), None
            )
        assert error_msg == str(excinfo.value)
        error_msg = ("The station longitude is required and must be a float or"
                     " int value between -180 and 180")
        tmp_dict = self.metadata._asdict()
        del tmp_dict["station_lon"]
        with pytest.raises(TypeError) as excinfo:
            check_required_parameters(
                "seasonal", ["median", "dark_aot40", "p90"],
                namedtuple("Metadata", tmp_dict)(**tmp_dict), None
            )
        assert error_msg == str(excinfo.value)
        error_msg = ("The station climatic zone is required and must be an int"
                     " or string contained in")
        tmp_dict = self.metadata._asdict()
        del tmp_dict["station_climatic_zone"]
        with pytest.raises(TypeError) as excinfo:
            check_required_parameters(
                "seasonal", ["median", "dark_aot40", "p90"],
                namedtuple("Metadata", tmp_dict)(**tmp_dict), None
            )
        assert str(excinfo.value).startswith(error_msg)


class TestCheckInputParameters:
    def test_check_input_parameters_monthly_drmdmax1h(self):
        error_msg = ("The drmdmax1h statistic cannot be evaluated with monthly"
                     " sampling")
        with pytest.raises(ValueError) as excinfo:
            check_input_parameters("monthly", "drmdmax1h", *[None]*10)
        assert error_msg == str(excinfo.value)

    def test_check_input_parameters_correct_input(self):
        data = pd.Series(
            range(10), index=pd.date_range(start="2000", periods=10, freq="H")
        )
        expected_required = namedtuple(
            "Required",
            "station_lat, station_lon, station_climatic_zone, seasons, crops"
        )(*[False]*5)
        result = check_input_parameters("annual", "median", data, *[None]*9)
        assert result.sampling == "annual"
        assert result.statistics == ["median"]
        pd.testing.assert_series_equal(result.data, data)
        assert result.metadata is None
        assert result.seasons is None
        assert result.crops is None
        assert result.min_data_capture is None
        assert result.required == expected_required
