"""Quantile regression calculation.

This module contains the following function:
quant_reg - calculate the quantile regression
"""

import numpy as np
import scipy.stats
import statsmodels.formula.api as smf

from toarstats.trends.utils import moving_block_bootstrap


def quant_reg(data, quantile):
    """Calculate the quantile regression.

    :param data: data containing a list of date time values and
                 associated parameter values on which to calculate the
                 trend
    :param quantile: a single quantile, must be between 0 and 1

    :return: The trend with its uncertainty and p value
    """
    fit = smf.quantreg("value~datetime", data).fit(q=quantile).params
    mbb = moving_block_bootstrap("quant", data, quantile)
    fit_se = np.nanstd(mbb, axis=0)
    fit_pv = 2*scipy.stats.t.sf(x=abs(fit/fit_se), df=len(data)-2)
    return {"trend": fit, "uncertainty": fit_se, "p_value": fit_pv}
